bl_info = {
  "name": "Quake Brush Map",
  "author": "Xembie",
  "version": (0, 0, 7),
  "blender": (2, 63, 0),
  "location": "File > Import-Export",
  "description": "Export Quake Brush Map",
  "warning": "",
  "wiki_url": "",
  "tracker_url": "",
  "category": "Import-Export"}

import math
class V:
	def __init__(self,array):
		self.array = array
	def __sub__(self,other):
		output = []
		for i in range(len(self.array)):
			output.append(self.array[i] - other.array[i])
		return V(output)
		
	def norm3d(self):
		t = self.array[0] + self.array[1] + self.array[2]
		if t == 0.0:
			return
		self.array[0] /= t
		self.array[1] /= t
		self.array[2] /= t
		
	def mag3d(self):
		return math.sqrt(	(self.array[0] * self.array[0]) + \
							(self.array[1] * self.array[1]) + \
							(self.array[2] * self.array[2]) )
	def dot3d(self,other):
		return	(self.array[0] * other.array[0]) + \
				(self.array[1] * other.array[1]) + \
				(self.array[2] * other.array[2])

def tri_normal(tri):
	a = V(tri[1])
	b = V(tri[2])
	a = a - V(tri[0])
	b = b - V(tri[0])
	normal = []
	normal.append((a.array[1] * b.array[2]) - (a.array[2] * b.array[1]))
	normal.append((a.array[2] * b.array[0]) - (a.array[0] * b.array[2]))
	normal.append((a.array[0] * b.array[1]) - (a.array[1] * b.array[0]))
	return normal
	
def tri_flip(tri):
	return [tri[0],tri[2],tri[1]]
	
def point_plane_side(normal,normalRef,point):
	x = V(normal)
	y = V(point)
	z = V(normalRef)
	if (x.dot3d(y) - x.dot3d(z)) > 0:
		return True
	return False
	
def plane_side(normal,ref,tri):
	if	point_plane_side(normal,ref,tri[0]) or \
		point_plane_side(normal,ref,tri[1]) or \
		point_plane_side(normal,ref,tri[2]):
		return True
	return False

def quake_brush_map(path,scale,texture):
	print("Quake Map exporter " + str(bl_info['version']) + " by Xembie:")
	objectlist = []
	brusheslist = []
	entitylist = []
	for obj in bpy.context.selected_objects:
		if obj.type == 'MESH':
			mesh = obj.to_mesh(bpy.context.scene,True,'PREVIEW')
			facelist = []
			invalidfaces = 0
			for f,face in enumerate(mesh.polygons):
				if len(face.vertices) < 3:
					invalidfaces += 1
					continue
				facelist.append([])
				for v in face.loop_indices:
					facelist[-1].append(mesh.loops[v].vertex_index)
			brushes = 0
			faces = 0
			concave = 0
			brush = []
			brusheslist.append(brush)#new entity
			propertylist = []
			brush.append(propertylist)
			brushlist = []
			brush.append(brushlist)
			objectlist.append(brush)
			propertylist.append([obj.name,"\"classname\" " + "\"" + obj.name + "\"\n"])
			for prop in obj.items():
				if prop[0][0] == '_':#skip underscored
					continue
				propertylist.append([str(prop[0]),"\"" + str(prop[0]) + "\" \"" + str(prop[1]) + "\"\n"])
			matrix = obj.matrix_world.to_4x4()
			matrix[0][1] *= -1.0
			matrix[0][3] *= -1.0
			matrix[1][0] *= -1.0
			while len(facelist) > 0:
				#add 1 face
				added = True
				writelist = []
				writelist.append(facelist[0])
				facelist.pop(0)
				#add all faces connected to faces in writelist and repeat until none are added.
				while added:
					added = False
					for i,face in enumerate(facelist):
						for f in writelist:
							if face[0] in f or face[1] in f or face[2] in f:
								added = True
								writelist.append(face)
								facelist.pop(i)
								break
				fliptri = []#index of tri to flip
				for x,face in enumerate(writelist):
					flipped = False
					for y,f in enumerate(writelist):
						if y == x:
							continue
						a = [[	mesh.vertices[f[0]].co[0],
								mesh.vertices[f[0]].co[1],
								mesh.vertices[f[0]].co[2] ],
							[	mesh.vertices[f[1]].co[0],
								mesh.vertices[f[1]].co[1],
								mesh.vertices[f[1]].co[2] ],
							[	mesh.vertices[f[2]].co[0],
								mesh.vertices[f[2]].co[1],
								mesh.vertices[f[2]].co[2] ]]
						normal = tri_normal(a)
						if plane_side(	normal,
										mesh.vertices[face[0]].co,
										a ):
							flipped = True
							break
					fliptri.append(flipped)
				#repeat to see if flipping the tri worked
				for x,face in enumerate(writelist):
					if fliptri[x]:
						for y,f in enumerate(writelist):
							if y == x:
								continue
							a = [[	mesh.vertices[f[0]].co[0],
									mesh.vertices[f[0]].co[1],
									mesh.vertices[f[0]].co[2] ],
								[	mesh.vertices[f[2]].co[0],
									mesh.vertices[f[2]].co[1],
									mesh.vertices[f[2]].co[2] ],
								[	mesh.vertices[f[1]].co[0],
									mesh.vertices[f[1]].co[1],
									mesh.vertices[f[1]].co[2] ]]
							normal = tri_normal(a)
							if plane_side(	normal,
											mesh.vertices[face[0]].co,
											a ):
								concave += 1
								break
				faces += len(writelist)
				if len(writelist) > 0:
					brushes += 1
					br = []
					brushlist.append(br)
					for x,face in enumerate(writelist):
						af = mesh.vertices[face[0]].co * matrix
						bf = mesh.vertices[face[1]].co * matrix
						cf = mesh.vertices[face[2]].co * matrix
						tri = [[	int(round((af[0] + obj.matrix_world[0][3]) * scale,0)),
									int(round((af[1] + obj.matrix_world[1][3]) * scale,0)),
									int(round((af[2] + obj.matrix_world[2][3]) * scale,0)) ],
								[	int(round((bf[0] + obj.matrix_world[0][3]) * scale,0)),
									int(round((bf[1] + obj.matrix_world[1][3]) * scale,0)),
									int(round((bf[2] + obj.matrix_world[2][3]) * scale,0)) ],
								[	int(round((cf[0] + obj.matrix_world[0][3]) * scale,0)),
									int(round((cf[1] + obj.matrix_world[1][3]) * scale,0)),
									int(round((cf[2] + obj.matrix_world[2][3]) * scale,0)) ]]
						if fliptri[x]:
							tri = tri_flip(tri)
						br.append(	"( "	+ str(tri[0][0]) + " " + str(tri[0][1]) + " " + str(tri[0][2]) + " ) " + \
									"( "	+ str(tri[1][0]) + " " + str(tri[1][1]) + " " + str(tri[1][2]) + " ) " + \
									"( "	+ str(tri[2][0]) + " " + str(tri[2][1]) + " " + str(tri[2][2]) + " ) " + \
									texture + "\n" )
			print(" Mesh(" + obj.name + ") " + str(brushes) + " brushes, " + str(faces) + " faces, " + str(invalidfaces) + " skipped, " + str(concave) + " concaves.") 
			bpy.data.meshes.remove(mesh)
		elif obj.type == 'EMPTY':#empties are entities
			entity = []
			entitylist.append(entity)#new entity
			propertylist = []
			entity.append(propertylist)
			origin = False
			for prop in obj.items():
				if prop[0][0] == '_':
					continue
				if prop[0] == 'origin':
					origin = True
				propertylist.append("\"" + str(prop[0]) + "\" \"" + str(prop[1]) + "\"\n")
			if not origin:
				propertylist.append("\"origin\" \"" +	str(int(round(obj.matrix_world[0][3] * scale,0))) + " " + \
														str(int(round(obj.matrix_world[1][3] * scale,0))) + " " + \
														str(int(round(obj.matrix_world[2][3] * scale,0))) + "\"\n")
			print(" Empty(" + obj.name + ") " + str(len(propertylist)) + " properties.")
	f = open(path,"wb")
	#write world with world properties
	f.write(b"{\n\"classname\" \"worldspawn\"\n")
	for prop in bpy.context.scene.world.items():
		if prop[0][0] == '_':
			continue
		f.write(("\"" + prop[0] + "\" \"" + prop[1] + "\"\n").encode('utf-8'))
	f.write(b"}\n")
	#Write objects
	for o in objectlist:
		f.write(b"{\n")
		for prop in o[0]:
			f.write((prop[1]).encode('utf-8'))
		for brush in o[1]:
			f.write(b"{\n")
			for br in brush:
				f.write((br).encode('utf-8'))
			f.write(b"}\n")
		f.write(b"}\n")
	#write empties
	if len(entitylist) > 0:
		for i,entity in enumerate(entitylist):
			f.write(b"{\n")
			for prop in entity[0]:
				f.write((prop).encode('utf-8'))
			f.write(b"}\n")
	f.close()
	print("Done")
	
from bpy.props import *
import bpy

class ExportBrushMap(bpy.types.Operator):
	bl_idname	= "export.brushmap"
	bl_label	= "Export brush map"
	filepath	= StringProperty(subtype = 'FILE_PATH',name="File path",description="Filepath for exporting",maxlen= 1024,default="brush.map")
	scale		= FloatProperty(name="Scale",description="Scale all objects from (0,0,0)",default=64.0,precision=5)
	texture		= StringProperty(name="Texture",description="Specific to some quake formats.",maxlen= 1024,default="common/caulk 0 0 0 1 1 0 0 0")
	def execute(self,context):
		quake_brush_map(self.properties.filepath,self.properties.scale,self.properties.texture)
		return {'FINISHED'}
	def invoke(self,context,event):
		wm = context.window_manager
		wm.fileselect_add(self)
		return {'RUNNING_MODAL'}

import bpy,os

def export(self,context):
	path = "brush.map"
	if len(bpy.data.filepath) > 0:
		path = os.path.splitext(bpy.data.filepath)[0] + ".map"
	self.layout.operator(ExportBrushMap.bl_idname,text="Quake Brush Map").filepath = path

def register():
	bpy.utils.register_module(__name__)
	bpy.types.INFO_MT_file_export.append(export)

def unregister():
	bpy.utils.unregister_module(__name__)
	bpy.types.INFO_MT_file_export.remove(export)

if __name__ == "__main__":
  register()
